/* ipaudit.c
 *
 * ipaudit - network traffic data gathering
 * By Jon Rifkin <j.rifkin@uconn.edu>
 * Copyright 1999 Jonathan Rifkin
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


/*
------------------------------------------------------------------------
Compile Switches
------------------------------------------------------------------------
*/
#define DEBUG

#define DUMP
#undef  DUMP

#define HASH
#undef  HASH


/*
------------------------------------------------------------------------
Include Files
------------------------------------------------------------------------
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <error.h>
#include <pcap.h>
#include <signal.h>
#include <sys/types.h>
#include <limits.h>
#include <unistd.h>
#include "hash.h"



/*
------------------------------------------------------------------------
Defines
------------------------------------------------------------------------
*/

#define TRUE 1
#define FALSE 0

#if 0
/*  TCP ports to dump  */
#define NPORT           6

#define P_FTP          21
#define P_TELNET       23
#define P_SUNRPC      111
#define P_LOGIN       513
#define P_INGRESLOCK 1524
#define P_IRC        6667

/*  Misc protocols to dump  */
#define PROT_ICMP       1
#endif

/*  Protocols with port info  */
#define PROT_TCP        6
#define PROT_UDP       17


/*  Flags for udp/tcp accepting all/some ports  */
#define PROT_ACC_ALL    1
#define PROT_ACC_SOME   2

/*  Length of saved packets  */
#define PLEN_DEF 96  /* default  */
#define PLEN_MIN 68  /* min allowed  */

/*  
Length of hash slots
NOT number of packets, they're unlimited (except for memory)
*/
#define NHSLOTS 1000000

/*
------------------------------------------------------------------------
DEBUGGING MACROS
------------------------------------------------------------------------
*/
#ifdef DEBUG
#define WRITEMSG \
   printf ("File %s line %d: ", __FILE__, __LINE__); \
   printf ("errmsg <%s>\n", strerror(errno)); fflush(stdout);
#define WRITEVAR(VAL,FMT) \
   printf ("File %s line %d: ", __FILE__, __LINE__); \
   printf ("%s=",#VAL); printf (#FMT, VAL); printf ("\n"); \
   fflush(stdout);
#else
#define WRITEMSG
#define WRITEVAR(VAL,FMT)
#endif


/*
------------------------------------------------------------------------
Type Definitions
------------------------------------------------------------------------
*/

typedef struct {
   u_char src[6];
   u_char dst[6];
   u_char ptype[2];     /*  ==0x800 if ip  */
   u_char version[1];
   u_char service[1];
   u_char length[2];
   u_char id[2];
   u_char flag[2];
   u_char ttl[1];
   u_char prot[1];
   u_char chksum[2];
   u_char srcip[4];
   u_char dstip[4];
   u_char srcpt[2];
   u_char dstpt[2];
   } pkt_struct_t;




/*
------------------------------------------------------------------------
Global variables
------------------------------------------------------------------------
*/
extern int errno;



/*
------------------------------------------------------------------------
Module variables
------------------------------------------------------------------------
*/
int isig_m=0;

u_char *prots_m = NULL;

u_char *tcp_ports_m = NULL;
u_char *udp_ports_m = NULL;


/*
------------------------------------------------------------------------
Local Function Prototypes
------------------------------------------------------------------------
*/
void ihandler (int);
void storepkt (pkt_struct_t *, hlist_t **);
void writepkt (hlist_t **, char *, char *);
void PrintUsage();

void parse_portstr(char *str);
void add_protocol(int val);
void add_port (int prot, int port);

void *alloc_err (int, int);
char *get_localstr (char *);



/*
------------------------------------------------------------------------
Main Function
------------------------------------------------------------------------
*/
int main (int argc, char *argv[]) {
   pcap_t *infile;
   char   ebuf[PCAP_ERRBUF_SIZE];
   struct pcap_pkthdr pkthdr;
   pkt_struct_t *pkt;
   struct bpf_program  fcode;
   pcap_dumper_t *df = NULL;
   int dump_this = FALSE;
   int nethpkt;
   int nippkt;
   int ival;
   int length;
   int i;
   u_char *dumptable = NULL;
   char *cmd = "ip";
   int  optchar;
   int  dumpall    = FALSE;
   FILE *pidfile   = NULL;
   char *progarg   = NULL;
   char *progfile  = NULL;
   char *writefile = NULL;
   char *readfile  = NULL;
   char *outfile   = NULL;
   int  maxpkt     = 0;
	int  nlen       = PLEN_DEF;   /*  Packet length to dump  */
	char *localstr  = NULL;			/*  Sorting order for ip address */
   /*  Hash table for ip connections  */
   hlist_t **hconn = NULL;
   



   /*  Read options  */
   while (-1 != (optchar=getopt(argc,argv,"ai:f:l:o:p:r:s:x:w:"))) {
      switch (optchar) {
      case '?':
         return 1;
      case 'a':
         dumpall = TRUE;
         break;
      /*  Write pid file  */
      case 'i':
         pidfile = fopen(optarg,"wt");
         if (NULL==pidfile) {
            fprintf (stderr, "ERROR:  Cannot open pidfile '%s'\n", optarg);
            exit(1);
         }
         fprintf (pidfile, "%d\n", getpid());
         fclose (pidfile);
         break;
      case 'x':
         progfile = strdup(optarg);
         break;
      case 'r':
         readfile = strdup(optarg);
         break;
      case 'w':
         writefile = strdup(optarg);
         break;
      case 'c':
         maxpkt    = atoi(optarg);
         break;
      case 'o':
         outfile   = strdup(optarg);
         break;
      case 'p':
         parse_portstr(strdup(optarg));
         break;
		case 's':
			nlen = atoi(optarg);
			if (nlen<PLEN_MIN)  
				nlen = PLEN_MIN;
			break;
		case 'l':
			localstr = get_localstr (optarg);
			break;
      default:
         return 1;
      }
   }
   
   
   
   /*  If not reading pcap file need the interface name  */
   if (!readfile && argc-optind<1)  {
      PrintUsage();
      return 0;
   }


   

   /*  Open pcap file  */
   if (readfile)
      infile = pcap_open_offline(readfile, ebuf);
   else {
      infile = pcap_open_live(argv[optind], nlen, 1, 1000, ebuf);
      if (infile==NULL) {
         printf("ipaudit: %s (%s)\n", ebuf, "Do you need root?");
         return 1;
      }
      if (pcap_compile(infile, &fcode, cmd, 0, 0) < 0)
         printf("compile: %s", pcap_geterr(infile));
      if (pcap_setfilter(infile, &fcode) < 0)
         printf("setfilter:  %s", pcap_geterr(infile));
   }



   /*  Install interupt handler if reading live */
   if (!readfile) {
      signal (SIGINT,    ihandler);   /*  intercepts  ^C           */
      signal (SIGTERM,   ihandler);   /*  intercepts  ^kill <PID>  */
   }

   /*  Initialize hash table  */
   hconn = init_htable(NHSLOTS);


   /*  Counters  */
   nethpkt = 0;
   nippkt = 0;

   /*  Open dump file  */
   if (writefile) df = pcap_dump_open(infile, writefile);

   /*  Read inteface until interupt signal  */
   while (isig_m==0) {
      nethpkt++;
      pkt = (pkt_struct_t *) pcap_next (infile, &pkthdr);

      /*  If pkt is null and we're reading file then we're done  */
      /*  Otherwise if reading live try again                    */
      if (pkt==NULL) {
         if (readfile) break;
         else          continue;
      }

      /*  Don't exceed limit of ip packets  */
      nippkt++;
      if (maxpkt && nippkt>maxpkt)
         break;

      /*  Set ports to 0 if not UDP or TCP  */
      if ( pkt->prot[0]!=0x11 && pkt->prot[0]!=0x06 ) {
         pkt->srcpt[0] = 0;
         pkt->srcpt[1] = 0;
         pkt->dstpt[0] = 0;
         pkt->dstpt[1] = 0;
      }

      /*  Store packets  */
      storepkt (pkt, hconn);


      /*  Using dump  */
      if (writefile) {

         /*  User specified to dump all packets  */
         if (dumpall) {
            dump_this = TRUE;

         } else {
            /*  See if this protocol is dumped  */
            dump_this = prots_m[pkt->prot[0]];

            /*  If udp or tcp, are ports specified ?  */
            if (PROT_ACC_SOME==dump_this) {
               dumptable = (PROT_TCP==pkt->prot[0]) ? tcp_ports_m : udp_ports_m;
               dump_this = 
                  dumptable[(pkt->srcpt[0]<<8)+pkt->srcpt[1]] ||
                  dumptable[(pkt->dstpt[0]<<8)+pkt->dstpt[1]];
            }
         }

         /*  Dump packet  */
         if (dump_this)
            pcap_dump ((u_char *) df, &pkthdr, (u_char *) pkt);
      }

      


   }  /*  Read tcpdump data  */


   /*  Close dump file  */
   if (writefile) pcap_dump_close (df);

   /*  Close packet input */
   pcap_close (infile);


   /*  Write packet info  */
   writepkt(hconn, outfile, localstr);

   /*  Call next program  */
   if (progfile) {
      progarg = index (progfile, ' ');
      if (progarg) {
         *progarg++ = '\0';
      }
      execl (progfile, progfile, progarg, NULL);
   }
}



/*
Interupt handler (called when program recieves operating system signal
*/
void ihandler (int cursig)
   {

   /*  SET SIGNAL FLAG  */
   isig_m = 1;

   /*  FLUSH BUFFERS  */
   fflush (stdout);

   /*  RE-INSTALL SIGNAL HANDLER  */
   signal (cursig, ihandler);
   }


/*
Store packet info in hash table, 
keyed by ip1,ip2,port1,por2,protocol
data  is number of incoming/outgoing bytes, packets
*/
void storepkt (pkt_struct_t *p, hlist_t **ha) {
   u_char key[13];
   int    idata[4];
   int    *data;
   int    ndata;
   int    length;

   /*  Calculate data packet length  */
   length = p->length[1] + 256*(int) p->length[0] + 14;

   /*  Make key - order so smallest ip first - store data  */
   if (memcmp(p->srcip, p->dstip, 4) < 0) {
      memcpy (key+ 0, p->srcip, 4);
      memcpy (key+ 4, p->dstip, 4);
      memcpy (key+ 8, p->srcpt, 2);
      memcpy (key+10, p->dstpt, 2);
      idata[0] = 0;
      idata[1] = length;
      idata[2] = 0;
      idata[3] = 1;
   } else {
      memcpy (key+ 0, p->dstip, 4);
      memcpy (key+ 4, p->srcip, 4);
      memcpy (key+ 8, p->dstpt, 2);
      memcpy (key+10, p->srcpt, 2);
      idata[0] = length;
      idata[1] = 0;
      idata[2] = 1;
      idata[3] = 0;
   }
   memcpy (key+12, p->prot,  1);

#ifdef DUMP
printf ("%03u.%03u.%03u.%03u ", key[0], key[1], key[2], key[3]);
printf ("%03u.%03u.%03u.%03u ", key[4], key[5], key[6], key[7]);
printf ("%u %u %u %d %d\n", key[12], (int )key[8]*256+key[9], (int )key[10]*256+key[11], length, 1);
#endif


   /*  Add to table  */
   if (! finddata (ha, (u_char *) &key, sizeof(key), (u_char **) &data, &ndata) ) {
      addnode (ha, (u_char *) &key, sizeof(key), (u_char *) idata, sizeof(idata));
   } else {
      data[0] += idata[0];
      data[1] += idata[1];
      data[2] += idata[2];
      data[3] += idata[3];

#ifdef DUMP
printf ("data idata  <%u %u %u %u>a  <%u %u %u %u>\n", 
data[0], data[1], data[2], data[3],
idata[0], idata[1], idata[2], idata[3]);
#endif
   }
}


/*
Retrieve and print packets from hash table 
*/
void writepkt (hlist_t **ha, char *outname, char *localstr) {
   hlist_t *t;
   int     *data;
   FILE    *outfile = stdout;
	char    ip1[16], ip2[16];
	int     pt1, pt2, prot;

   if (outname) 
      outfile = fopen (outname, "wt");

   if (NULL==outfile) {
      fprintf (stderr, "ERROR:  Cannot open output file <%s>\n", outname);
      exit(1);
   }

   /*  Walk list  */
   getfirst (ha);
   while ((t=getnext(ha))) {

		/*  Get ip addresses and ports  */
		sprintf (ip1, "%03u.%03u.%03u.%03u", 
			t->key[0], t->key[1], t->key[2], t->key[3]);
		sprintf (ip2, "%03u.%03u.%03u.%03u", 
			t->key[4], t->key[5], t->key[6], t->key[7]);
      pt1  = (int) t->key[ 8]*256 + t->key[ 9];
      pt2  = (int) t->key[10]*256 + t->key[11];
		prot = t->key[12];

		/*  Re-order ip addresses if 2nd is local and first is not  */
		if (
			localstr && 
		 	 strncmp(localstr,ip1,strlen(localstr)) &&
			!strncmp(localstr,ip2,strlen(localstr)) ) {
		
			/*  Print key info  */
			fprintf (outfile, "%s %s %u %u %u", ip2, ip1, prot, pt2, pt1);

			/*  Data  */
			data = (int *) t->data;
			fprintf (outfile, "  %d %d %d %d\n", 
				data[1], data[0], data[3], data[2]); 

		} else {
		
			/*  Print key info  */
			fprintf (outfile, "%s %s %u %u %u", ip1, ip2, prot, pt1, pt2);

			/*  Data  */
			data = (int *) t->data;
			fprintf (outfile, "  %d %d %d %d\n", 
				data[0], data[1], data[2], data[3]); 
		}
#if 0
      /*  IP addresses  */
      fprintf (outfile, "%03u.%03u.%03u.%03u ", 
			t->key[0], t->key[1], t->key[2], t->key[3]);
      fprintf (outfile, "%03u.%03u.%03u.%03u ", 
			t->key[4], t->key[5], t->key[6], t->key[7]);
      /*  Protocol  */
      fprintf (outfile, "%u ", t->key[12]);
      /*  Ports      */
      fprintf (outfile, "%u ", (int) t->key[ 8]*256 + t->key[ 9]);
      fprintf (outfile, "%u ", (int) t->key[10]*256 + t->key[11]);
#endif
   }

   /*  Close file  */
   if (outname)
      fclose(outfile);
   
};


void PrintUsage(void) {
   printf ("\nUsage: ipaudit -acioprwx [interface]\n");
   printf ("  Read and record info on ip connections and optionally\n");
   printf ("  dump packets to file\n");
   printf ("\n");
   printf ("  -r readfile   -  Read packets from pcap format file\n");
   printf ("                   Don't need interface with this option\n");
   printf ("  -w writefile  -  Dump selected packets to pcap format file\n");
   printf ("                   MUST also supply either -a or -p options\n");
   printf ("  -a            -  Dump ALL packets to dumpfile.\n");
   printf ("                   Incompatible with -p option\n");
   printf ("  -i pidfile    -  Write process id to file\n");
	printf ("  -l net-ip     -  Order output ip address pairs by network ip\n");
   printf ("  -s nlen       -  Dump first <nlen> bytes of each packet (default %d, min %d)\n",
		PLEN_DEF, PLEN_MIN);
   printf ("  -x program    -  Run program when done\n");
   printf ("  -c npacket    -  Only read in specific number of ip packets\n");
   printf ("  -o outfile    -  Place output in 'outfile'. Default is stdout\n");
   printf ("  -p string     -  Dump only selected ip protocols and ports\n");
   printf ("     string format -p n:n:n,p,p:n where n is protocol number\n");
   printf ("     and p is port number (only for protocols 6 (tcp) and 17 (udp)\n");
   printf ("     Incompatible with -a option.\n");
   printf ("Example:\n");
   printf ("  ipaudit -w dump.fil -p1:2:6,21,23  eth0\n");
   printf ("Write only packets with protocols 1 (icmp), 2 (?), and 6 (tcp)\n");
   printf ("and tcp ports 21 and 23 (ftp,telnet)\n\n");
}



/*
Determine protocols and ports to accept from string of format
   "1:6,21,23,6667:17" means protocols 1 (icmp), 6 (tcp) and 17 (udp)
   For tcp only ports 21,23,6667.
   For udp all ports (default)
*/
void parse_portstr(char *str) {
   char *p;
   char delim;
   int  lastprot = -1;
   int  is_prot;
   int  val;

   is_prot = TRUE;
   delim=' ';
   while (delim) {
      /*  Find : or '\0'  */
      p=str;
      /*  Find next : , '\0'  */
      while (*p!=':' && *p!=',' && *p!='\0') 
         p++;
      delim = *p;
      *p = '\0';
      val = atoi(str);
      str = p+1;

      if (is_prot) {
         lastprot = val;
         add_protocol(val);
      } else {
         add_port (lastprot,val);
      }
      is_prot = (delim==':');
   }
}


void add_protocol(int val) {
   /*  Check for valid protocol  */
   if (val<0 || val>255)
      return;
   /*  ALlocate if not done already  */
   if (NULL==prots_m)
      prots_m = (u_char *) alloc_err(256, sizeof(u_char));
   if (NULL==prots_m) {
      fprintf (stderr, "ERROR: Out of memory\n");
      exit (1);
   }
   /* Accept all ports  */
   prots_m[val] = PROT_ACC_ALL;
   }


void add_port (int prot, int port) {
   /*  Check for valid port  */
   if (prot<0 || prot>0x0ffff)
      return;
   /*  Check for valid protocol  */
   if (prot==PROT_TCP) {
      prots_m[prot] = PROT_ACC_SOME;
      if (NULL==tcp_ports_m) 
         tcp_ports_m = (u_char *) alloc_err(0x010000, sizeof(u_char));
      tcp_ports_m[port] = 1;
   } else if (prot==PROT_UDP) {
      prots_m[prot] = PROT_ACC_SOME;
      if (NULL==udp_ports_m) 
         udp_ports_m = (u_char *) alloc_err(0x010000, sizeof(u_char));
      tcp_ports_m[port] = 1;
   }
}



void *alloc_err (int n, int size) {
   void *p;
   p = calloc (n, size);
   if (NULL==p) {
      fprintf (stderr, "ERROR: Out of memory.\n");
      exit(1);
   }
   return p;
}


/*  
Read and format ip address :
  for example,  "112.51.1" -> "112.051.001"
*/
char *get_localstr(char *instr) {
   /*  Room for string containing formated ip address  */
	char *buffer = (char *) calloc(16,sizeof(char)); 
	char *b = buffer;
	char triplet[3];
	char *p = instr;
	int  i,n;

	/*  Search over all bytes (things between ..)  */
	while (1) {
		n = 0;
		while ((*p>='0' && *p<='9') && n<3) 
			triplet[n++] = *p++;
		/*  Either non-digit or two many digits - return NULL */
		if (*p!='.' && *p!=0) return NULL;
		for (i=n;i<3;i++) *b++ = '0';
		for (i=0;i<n;i++) *b++ = triplet[i];
		*b++ = *p;
		if (*p==0)	return buffer;
		p++;
	}
}

